@@ -1,4 +1,5 @@
 use logos::{Lexer, Logos};
+use std::collections::HashSet;
 
 #[derive(Logos, Debug, PartialEq)]
 enum Token {
@@ -20,11 +21,14 @@ enum Token {
     #[regex("any of")]
     AnyExpression,
 
-    #[regex(r#""(\\"|[^"\n])*""#, raw)]
-    RawDouble(String),
+    #[regex(r#""(\\"|[^"\n])*""#, literal)]
+    LiteralDouble(String),
 
-    #[regex(r#"'(\\'|[^'\n])*'"#, raw)]
-    RawSingle(String),
+    #[regex(r#"'(\\'|[^'\n])*'"#, literal)]
+    LiteralSingle(String),
+
+    #[regex(r#"`(\\`|[^`\n])*`"#, raw)]
+    Raw(String),
 
     #[regex("[a-z] to [a-z]", range)]
     LowercaseRange(String),
@@ -35,7 +39,7 @@ enum Token {
     #[regex("[0-9] to [0-9]", range)]
     NumericRange(String),
 
-    #[regex(r#"capture [a-z]+ \{"#, named_capture)]
+    #[regex(r#"capture \w+ \{"#, named_capture)]
     NamedCapture(String),
 
     #[token("capture {")]
@@ -47,11 +51,14 @@ enum Token {
     #[token("either {")]
     Either,
 
-    #[token("start")]
-    LineStart,
+    #[token("}")]
+    BlockEnd,
 
-    #[token("end")]
-    LineEnd,
+    #[token("<start>")]
+    StartSymbol,
+
+    #[token("<end>")]
+    EndSymbol,
 
     #[token("<newline>")]
     NewlineSymbol,
@@ -89,11 +96,8 @@ enum Token {
     #[token("<vertical>")]
     VerticalSymbol,
 
-    #[token("char")]
-    Char,
-
-    #[token("}")]
-    GroupEnd,
+    #[token("<char>")]
+    CharSymbol,
 
     #[token(";")]
     Semicolon,
@@ -112,6 +116,7 @@ enum Token {
 enum QuoteType {
     Single,
     Double,
+    Raw,
 }
 
 #[derive(Debug, Clone)]
@@ -164,22 +169,49 @@ fn get_quote_type(quote: &str) -> QuoteType {
     }
 }
 
-fn escape_quotes(source: String, quote_type: QuoteType) -> String {
+fn unescape_quotes(source: String, quote_type: QuoteType) -> String {
     match quote_type {
         QuoteType::Double => source.replace(r#"\""#, r#"""#),
         QuoteType::Single => source.replace(r#"\'"#, r#"'"#),
+        QuoteType::Raw => source.replace(r#"\`"#, r#"`"#),
     }
 }
 
-fn remove_and_escape_quotes(source: &str) -> String {
+fn escape_chars(source: String) -> String {
+    let reserved_chars = HashSet::from([
+        '[', ']', '(', ')', '{', '}', '*', '+', '?', '|', '^', '$', '.', '-', '\\',
+    ]);
+    let mut escaped_source = String::new();
+    for char in source.chars() {
+        if reserved_chars.contains(&char) {
+            let escaped_char = format!("\\{char}");
+            escaped_source.push_str(&escaped_char);
+        } else {
+            escaped_source.push_str(&String::from(char))
+        }
+    }
+    escaped_source
+}
+
+fn escape(source: String, quote_type: QuoteType) -> String {
+    escape_chars(unescape_quotes(source, quote_type))
+}
+
+fn remove_and_escape(source: &str) -> String {
     let pattern = source[1..source.len() - 1].to_owned();
     let quote = source[0..1].to_owned();
     let quote_type = get_quote_type(&quote);
-    escape_quotes(pattern, quote_type)
+    escape(pattern, quote_type)
+}
+
+fn literal(lex: &mut Lexer<Token>) -> String {
+    remove_and_escape(lex.slice())
 }
 
 fn raw(lex: &mut Lexer<Token>) -> String {
-    remove_and_escape_quotes(lex.slice())
+    let source = lex.slice();
+    let pattern = source[1..source.len() - 1].to_owned();
+    unescape_quotes(pattern, QuoteType::Raw)
 }
 
 fn handle_quantifier(source: String, quantifier: Option<String>, group: bool) -> Option<String> {
@@ -209,10 +241,6 @@ fn create_parse_error(lex: &mut Lexer<Token>, line: u16) -> ParseError {
     }
 }
 
-fn in_block(in_group: bool, in_either: bool) -> bool {
-    in_group || in_either
-}
-
 /**
 Compiles Melody source code to a regular expression
 
@@ -242,14 +270,16 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
 
     let mut group_quantifier = None;
 
+    let mut store: Vec<String> = Vec::new();
+
     let mut output = String::new();
 
-    let mut stack: Vec<String> = Vec::new();
-
     while let Some(token) = lex.next() {
+        let in_block = in_group || in_either;
+
         let formatted_token = match token {
             // raw
-            Token::RawDouble(pattern) | Token::RawSingle(pattern) => {
+            Token::LiteralDouble(pattern) | Token::LiteralSingle(pattern) | Token::Raw(pattern) => {
                 let group = pattern.chars().count() != 1;
                 handle_quantifier(pattern, quantifier.clone(), group)
             }
@@ -261,7 +291,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
 
             // groups
             Token::Capture => {
-                if in_block(in_group, in_either) {
+                if in_block {
                     return Err(create_parse_error(&mut lex, line));
                 }
                 group_quantifier = quantifier;
@@ -270,7 +300,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
                 Some(String::from("("))
             }
             Token::NamedCapture(name) => {
-                if in_block(in_group, in_either) {
+                if in_block {
                     return Err(create_parse_error(&mut lex, line));
                 }
                 group_quantifier = quantifier;
@@ -279,7 +309,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
                 Some(format!("(?<{name}>"))
             }
             Token::Match => {
-                if in_block(in_group, in_either) {
+                if in_block {
                     return Err(create_parse_error(&mut lex, line));
                 }
                 group_quantifier = quantifier;
@@ -288,7 +318,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
                 Some(String::from("(?:"))
             }
             Token::Either => {
-                if in_block(in_group, in_either) {
+                if in_block {
                     return Err(create_parse_error(&mut lex, line));
                 }
                 group_quantifier = quantifier;
@@ -296,7 +326,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
                 in_either = true;
                 None
             }
-            Token::GroupEnd => {
+            Token::BlockEnd => {
                 if in_group {
                     in_group = false;
                     let current_group_quantifier = group_quantifier;
@@ -304,12 +334,12 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
                     handle_quantifier(String::from(")"), current_group_quantifier, false)
                 } else if in_either {
                     in_either = false;
-                    let inner_expressions = stack.join("|");
-                    stack.clear();
+                    let inner_expressions = store.join("|");
+                    store.clear();
                     let current_group_quantifier = group_quantifier;
                     group_quantifier = None;
                     handle_quantifier(
-                        format!("({})", inner_expressions),
+                        format!("(?:{})", inner_expressions),
                         current_group_quantifier,
                         false,
                     )
@@ -345,8 +375,8 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
             }
 
             // direct replacements
-            Token::LineStart => handle_quantifier(String::from("^"), quantifier.clone(), false),
-            Token::LineEnd => handle_quantifier(String::from("$"), quantifier.clone(), false),
+            Token::StartSymbol => handle_quantifier(String::from("^"), quantifier.clone(), false),
+            Token::EndSymbol => handle_quantifier(String::from("$"), quantifier.clone(), false),
             Token::SpaceSymbol => handle_quantifier(String::from("\\s"), quantifier.clone(), false),
             Token::NotSpaceSymbol => {
                 handle_quantifier(String::from("\\S"), quantifier.clone(), false)
@@ -371,7 +401,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
             Token::VerticalSymbol => {
                 handle_quantifier(String::from("\\v"), quantifier.clone(), false)
             }
-            Token::Char => handle_quantifier(String::from("."), quantifier.clone(), false),
+            Token::CharSymbol => handle_quantifier(String::from("."), quantifier.clone(), false),
 
             // warning and error related
             Token::NewLine => {
@@ -383,7 +413,7 @@ pub fn compiler(source: &str) -> Result<String, ParseError> {
 
         if let Some(formatted_token) = formatted_token {
             if in_either {
-                stack.push(formatted_token);
+                store.push(formatted_token);
             } else {
                 output.push_str(&formatted_token);
             }
@@ -480,9 +510,9 @@ fn open_range_expression_test() {
 fn start_end_test() {
     let output = compiler(
         r#"
-      start;
+      <start>;
       "a"
-      end;
+      <end>;
       "#,
     )
     .unwrap();
@@ -493,6 +523,8 @@ fn start_end_test() {
 fn symbol_test() {
     let output = compiler(
         r#"
+      <start>;
+      <char>;
       <space>;
       not <space>;
       <newline>;
@@ -505,10 +537,11 @@ fn symbol_test() {
       <word>;
       not <word>;
       <vertical>;
+      <end>;
       "#,
     )
     .unwrap();
-    assert_eq!(output, r"/\s\S\n\t\r\f\0\d\D\w\W\v/");
+    assert_eq!(output, r"/^.\s\S\n\t\r\f\0\d\D\w\W\v$/");
 }
 
 #[test]
@@ -552,7 +585,7 @@ fn comment_test() {
 fn char_test() {
     let output = compiler(
         r#"
-      3 of char;
+      3 of <char>;
       "#,
     )
     .unwrap();
@@ -563,7 +596,7 @@ fn char_test() {
 fn some_test() {
     let single_output = compiler(
         r#"
-      some of char;
+      some of <char>;
       "#,
     )
     .unwrap();
@@ -581,7 +614,7 @@ fn some_test() {
 fn option_test() {
     let single_output = compiler(
         r#"
-      option of char;
+      option of <char>;
       "#,
     )
     .unwrap();
@@ -611,14 +644,14 @@ fn either_test() {
       "#,
     )
     .unwrap();
-    assert_eq!(output, "/(first|second|[a-z])(first|second)/");
+    assert_eq!(output, "/(?:first|second|[a-z])(?:first|second)/");
 }
 
 #[test]
 fn any_test() {
     let single_output = compiler(
         r#"
-      any of char;
+      any of <char>;
       "#,
     )
     .unwrap();
@@ -631,3 +664,14 @@ fn any_test() {
     .unwrap();
     assert_eq!(multiple_output, "/(?:ABC)*/");
 }
+
+#[test]
+fn raw_test() {
+    let output = compiler(
+        r#"
+      5 of `.*`
+      "#,
+    )
+    .unwrap();
+    assert_eq!(output, "/(?:.*){5}/");
+}
